% % % Performs particle swarm model selection and returns the 
% % % selected model as an object
% % % 
% % % 
% % % Input:
% % %           retAlgo  -  psms - object
% % %           dat      -  training data
% % % Output:
% % %           my_models -         a struct with the following two fields
% % %           my_models.retRes    The best model obtained throught the
% % %                               search process.
% % %           my_models.Best_set  the set of the best k-models found
% % %                               throught the search process. (k = psms.save_every)
% % %           Best_set -  the set of best particles found throught the
% % %                       search
% % %           retAlgo  -  the best model found throught the psms
% % % 
% % % If would you like to include other methods for the search process the
% % % following scripts should be modified:
% % %     psms (change the dimensionality a.Dim in psms.m)
% % %     validate_particle.m
% % %     get_prep_pars.m
% % %     get_classifier_pars.m
% % % 
% % %===============================================================
% % % Hugo Jair Escalante -- hugojair@ccc.inaoep.mx -- November 2007
% % %                        hugo.jair@gmail.com
% % %===============================================================

function [dat, retRes] = training(retAlgo,dat)
rand('state',sum(100*clock));

% % % Here we should be able of modifying the particles representation for
% % % allowing model selection of a given child model
if (~isclop((retAlgo.child{1})) || retAlgo.fms==1)
    retAlgo.fms=1;   %%% Perform full model selection
% % %  In this case we will perform full model selection
else
% retAlgo.fms=0;
untrained = retAlgo.child{1};  %%% The model we would like to optimize
base_algo=untrained.child{1,1}.name;
if isa(untrained,'cell') 
    untrained = group(untrained); % <-- convert retAlgo.child{1} to group
end;
end;
history_best=-1.*ones(1,retAlgo.SwarmSize*retAlgo.Iterations+retAlgo.SwarmSize);
[num,vecDim,outDim]=get_dim(dat);       %% Dimensionality of data
orig_name=get_name(dat);                
retRes = {};
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      PSMS STARTS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initializing variables
success = 0; % Success Flag
iter = 0;   % Iterations' counter

% % Defines total number of iterations for which weight is changed.
w_varyfor = floor(retAlgo.w_varyfor*retAlgo.Iterations); %Weight change step. 
w_now = retAlgo.w_start;
inertdec = (retAlgo.w_start-retAlgo.w_end)/w_varyfor; %Inertia weight's change per iteration
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%   SWARM INITIALIZATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% % % % LOWER AND UPPER BOUNDS FOR THE PARTICLES REPRESENTATION 
% % % %         (preprocessing and feature selection)
[datasets.lbs datasets.ubs]=get_prep_pars(retAlgo,vecDim,num);

for i=1:retAlgo.SwarmSize
    for j=1:length(datasets.lbs)
        preSwarm(i,j) = rand*(datasets.ubs(j)-datasets.lbs(j)) + datasets.lbs(j);
    end
    if (retAlgo.fms==1)
    rn=rand;
    [classifier]=get_classifier(retAlgo,rn);
    else
        [classifier,rn]=get_classifier_clop(retAlgo,base_algo);
    end
    [classifier datasets.classifierlb datasets.classifierub]=get_classifier_pars(retAlgo,classifier);
    for j=1:length(datasets.classifierlb)
        preSwarm_2(i,j) = rand*(datasets.classifierub(j)-datasets.classifierlb(j)) + datasets.classifierlb(j);
    end            
        Swarm(i,:)=[preSwarm(i,:), rn, preSwarm_2(i,:)];
end 
% % % End of swarm initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

VStep = rand(retAlgo.SwarmSize, retAlgo.Dim);
for hh = 1:retAlgo.SwarmSize
    [Swarm(hh,:), fSwarm(hh)] =  fitness(retAlgo,dat,Swarm(hh,:),vecDim);
    track(hh,:)=fSwarm(hh);
end

% Initializing the Best positions matrix and
% the corresponding function values
PBest = Swarm;
fPBest = fSwarm;

% Finding best particle in initial population
[fGBest, g] = min(fSwarm);
lastbpf = fGBest;
Best = Swarm(g,:); %Used to keep track of the Best particle ever
Best_set=Best;
fBest = fGBest;
history = [0, fGBest];
counter=hh+1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                  THE  PSO  LOOP                          %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
while( (success == 0) & (iter < retAlgo.Iterations) )
    iter = iter+1;
    fprintf('************ Iteration # %i ******************\n',iter);    
    str=[' Iteration=' num2str(iter)];      
    
    % Update the value of the inertia weight W
        if (iter<=w_varyfor) & (iter > 1)
        w_now = w_now - inertdec; %Change inertia weight
    end
           
    % Set GBest
    A = repmat(Swarm(g,:), retAlgo.SwarmSize, 1); %A = GBest. repmat(X, m, n) repeats the matrix X in m rows by n columns.
    B = A; %B wil be nBest (best neighbor) matrix
        
    % Generate Random Numbers
    R1 = rand(retAlgo.SwarmSize, retAlgo.Dim);
    R2 = rand(retAlgo.SwarmSize, retAlgo.Dim);
    
    % Calculate Velocity
    VStep = w_now*VStep + retAlgo.c1*R1.*(PBest-Swarm) + retAlgo.c2*R2.*(A-Swarm);
        
    % Apply Vmax Operator for v > Vmax
    changeRows = VStep > retAlgo.Vmax;
    VStep(find(changeRows)) = retAlgo.Vmax;
    % Apply Vmax Operator for v < -Vmax
    changeRows = VStep < -retAlgo.Vmax;
    VStep(find(changeRows)) = -retAlgo.Vmax;
    
    % ::UPDATE POSITIONS OF PARTICLES::
    Swarm = Swarm + retAlgo.Chi * VStep;    % Evaluate new Swarm    

    % % %  If not full-model selection, we should keep the classifier
    % % %   specified by the user
    if (retAlgo.fms~=1)
		Swarm(:,20)=rn;
	end
    for hh = 1:retAlgo.SwarmSize
        [Swarm(hh,:), fSwarm(hh)] =  fitness(retAlgo,dat,Swarm(hh,:),vecDim);
        track(counter,:)=fSwarm(hh);
        counter=counter+1;
    end


    % Updating the best position for each particle
    changeRows = fSwarm < fPBest;
    fPBest(find(changeRows)) = fSwarm(find(changeRows));
    PBest(find(changeRows), :) = Swarm(find(changeRows), :);
    
    % Updating index g
    [fGBest, g] = min(fPBest);
    
    %Update Best. Only if fitness has improved.   
    if fGBest < lastbpf
        [fBest, b] = min(fPBest);
        Best = PBest(b,:);
    end
    
    [valid_particle] = validate_particle(retAlgo,Best, vecDim);
    [retRes,base_model]=create_CLOP_model(retAlgo,valid_particle);
    
    if (mod(iter,retAlgo.save_every)==0)
        Best_set=[Best_set; Best];       
        history_best(:,iter*retAlgo.SwarmSize+retAlgo.SwarmSize)=fBest;
    end
    save psms_bestmodels.mat retRes Best_set;
    retRes = {};
    %%TERMINATION%%
    if abs(fGBest-0) <= retAlgo.TerminateErr     %GBest
        success = 1;
    elseif abs(fBest-0)<=retAlgo.TerminateErr    %Best
        success = 1;
    else
        lastbpf = fGBest; %To be used to find Best
    end
    
    if retAlgo.algorithm.verbosity
        str = sprintf('PSMS, iteration No. %d of a total of %d', iter, retAlgo.Iterations);
        disp(str)
    end  
    
% if retAlgo.visualize
% plot(track,'g','LineWidth',2,'Marker','*','MarkerEdgeColor','r');set(figure(1),'Color','w');
% figure(1); hold on; stem(history_best,'MarkerEdgeColor','r');
% % figure(1); hold on; plot(history_best,'.','MarkerEdgeColor','b');plot(history_best,'o','MarkerEdgeColor','r');
% title('Fitness function evaluation');
% xlabel('Particles fitness value');ylabel('CV-BER');
% xlim([0 size(track,1)+1]);
% ylim([(min(track)-0.15) (max(track)+0.15)]);
% legend('fitness per particle per iteration','best fitness per iteration');
% % pause(5);close all;    
% end

end
% % % In case we should return an ensemble
if (retAlgo.create_ensemble==1)
clear retRes;
for i=1:size(Best_set,1)
    [valid_particle] = validate_particle(retAlgo,Best_set(i,:), vecDim);
    [members{i},base_model]=create_CLOP_model(retAlgo,valid_particle);
    clear valid_particle;
end
   retRes=ensemble(members);
% % %    An ensemble with the final swarm
elseif (retAlgo.create_ensemble==2)
clear retRes;
for i=1:retAlgo.SwarmSize
    [valid_particle] = validate_particle(retAlgo,Swarm(i,:), vecDim);
    [members{i},base_model]=create_CLOP_model(retAlgo,valid_particle);
    clear valid_particle;
end
   retRes=ensemble(members);
   
% % % Or the best model
else
    [valid_particle] = validate_particle(retAlgo,Best, vecDim);
    [retRes,base_model]=create_CLOP_model(retAlgo,valid_particle);
    
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                  END  OF PSO  LOOP                       %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if retAlgo.visualize
plot(track,'g','LineWidth',2,'Marker','*','MarkerEdgeColor','r');set(figure(1),'Color','w');
figure(1); hold on; stem(history_best,'MarkerEdgeColor','r');
% figure(1); hold on; plot(history_best,'.','MarkerEdgeColor','b');plot(history_best,'o','MarkerEdgeColor','r');
title('Fitness function evaluation');
xlabel('Particles fitness value');ylabel('CV-BER');
xlim([0 size(track,1)+1]);
ylim([(min(track)-0.15) (max(track)+0.15)]);
legend('fitness per particle per iteration','best fitness per iteration');
% pause(5);close all;    
end
